@@ -4,20 +4,30 @@ Functions for working with PKDB data.
 * FIXME: specify which methods modify or copy data frames
 """
 import logging
-import warnings
+import os
+import tempfile
+import zipfile
 from abc import ABC
+from ast import literal_eval
 from collections import OrderedDict
-from copy import copy
+from io import BytesIO
 from pathlib import Path
-from typing import Callable, List
+from typing import Callable, Dict, Iterable, List, Union
 
 import numpy as np
 import pandas as pd
+import requests
 from IPython.display import display
-from pandas.errors import PerformanceWarning
+
+from pkdb_analysis.filter import f_healthy, f_n_healthy, filter_factory
+from pkdb_analysis.units import ureg
+from pkdb_analysis.utils import create_parent
 
 
-warnings.simplefilter(action="ignore", category=PerformanceWarning)
+# from pandas.errors import PerformanceWarning
+# This is not fixing anything, but just ignoring the problem !!!
+# warnings.simplefilter(action="ignore", category=PerformanceWarning)
+
 logger = logging.getLogger(__name__)
 
 
@@ -29,18 +39,19 @@ class PKDataFrame(pd.DataFrame, ABC):
 
     @property
     def _constructor(self):
+        """Internal function need for inheritances from pd.DataFrame."""
         return PKDataFrame._internal_ctor
 
     _metadata = ["pk"]
 
     @classmethod
     def _internal_ctor(cls, *args, **kwargs):
+        """Internal function need for inheritances from pd.DataFrame."""
         kwargs["pk"] = None
         return cls(*args, **kwargs)
 
     def __init__(self, data, pk=None, index=None, columns=None, dtype=None, copy=True):
         """
-
         :param data:
         :param pk:
         :param index:
@@ -57,7 +68,8 @@ class PKDataFrame(pd.DataFrame, ABC):
         self.pk = pk
 
     @staticmethod
-    def _validate_not_in_columns(columns):
+    def _validate_not_in_columns(columns: Iterable[str]):
+        """ """
         if columns:
             assert "pk" in columns
 
@@ -96,13 +108,46 @@ class PKDataFrame(pd.DataFrame, ABC):
         df_excluded = self[~self[self.pk].isin(df_pks)]
         return PKDataFrame(df_excluded, pk=self.pk)
 
+    @staticmethod
+    def _change_unit(sd, unit):
+        infer_fields = ["value", "mean", "median", "min", "max", "sd", "se"]
+        return PKDataFrame._change_unit_generic(
+            sd=sd, unit=unit, infer_fields=infer_fields, unit_field="unit"
+        )
+
+    @staticmethod
+    def _change_unit_generic(sd, unit: str, infer_fields: List[str], unit_field: str):
+        if isinstance(sd[unit_field], str):
+            if ureg(sd[unit_field]).check(unit):
+                factor = ureg(sd[unit_field]).to(unit).m
+                sd[infer_fields] = sd[infer_fields] * factor
+                sd[unit_field] = unit
+        return sd
+
+    @staticmethod
+    def _change_time_unit(sd, unit):
+        infer_fields = ["time"]
+        unit_field = "time_unit"
+        return PKDataFrame._change_unit_generic(
+            sd=sd, unit=unit, infer_fields=infer_fields, unit_field=unit_field
+        )
+
+    def change_unit(self, unit):
+        df = self.df.apply(self._change_unit, unit=unit, axis=1)
+        return PKDataFrame(df, pk=self.pk)
+
+    def change_time_unit(self, unit):
+        df = self.df.apply(self._change_time_unit, unit=unit, axis=1)
+        return PKDataFrame(df, pk=self.pk)
+
     @property
     def pk_column(self):
+        """Returns the column containing the primary value of this table"""
         return self[self.pk]
 
     @property
     def pks(self) -> set:
-        """ Set of pks."""
+        """Set of pks."""
         if self.pk in self.df.columns:
             return set(self[self.pk].unique())
         else:
@@ -110,24 +155,27 @@ class PKDataFrame(pd.DataFrame, ABC):
 
     @property
     def pk_len(self) -> int:
+        """returns number of unique identifiers within the table. This value can be smaller than the number
+        of rows. Since multiple rows can represent one instances."""
         return len(self.pks)
 
     @property
     def df(self) -> pd.DataFrame:
-        """ Returns a copied DataFrame."""
+        """Returns a copied DataFrame."""
         df = self.copy()
         del df.pk
         return pd.DataFrame(self)
 
     @property
     def study_sids(self) -> set:
-        """ Set of study_sids."""
+        """Set of study_sids."""
         study_sids = set([])
         if "study_sid" in self.df.columns:
             study_sids = set(self.study_sid.unique())
         return study_sids
 
     def _emptify(self) -> "PKDataFrame":
+        """Removes all entries from table."""
         empty_df = pd.DataFrame([], columns=self.columns)
         return PKDataFrame(empty_df, pk=self.pk)
 
@@ -142,15 +190,7 @@ class PKDataFrame(pd.DataFrame, ABC):
 
 
 class PKData(object):
-    PK_COLUMNS = {
-        "studies": "study_pk",
-        "groups": "group_pk",
-        "individuals": "individual_pk",
-        "interventions": "intervention_pk",
-        "outputs": "output_pk",
-        "data": "subset_pk",
-    }
-    """ Consistent set of data from PK-DB.set -a && source .env.local
+    """Consistent set of data from PK-DB.set -a && source .env.local
 
     Information is stored as DataFrames.
 
@@ -160,18 +200,40 @@ class PKData(object):
     - individuals
     - interventions
     - outputs
-    - data
+    - timecourses
     """
-    KEYS = ["studies", "groups", "individuals", "interventions", "outputs", "data"]
+
+    PK_COLUMNS = {
+        "studies": "study_pk",
+        "groups": "group_pk",
+        "individuals": "individual_pk",
+        "interventions": "intervention_pk",
+        "outputs": "output_pk",
+        "timecourses": "subset_pk",
+        "scatters": "subset_pk",
+    }
+
+    KEYS = [
+        "studies",
+        "groups",
+        "individuals",
+        "interventions",
+        "outputs",
+        "timecourses",
+        "scatters",
+    ]
+
+    # PK_COLUMNS = {key: f"{key[:-1]}_pk" for key in KEYS}
 
     def __init__(
         self,
+        studies: pd.DataFrame = None,
         interventions: pd.DataFrame = None,
         groups: pd.DataFrame = None,
         individuals: pd.DataFrame = None,
         outputs: pd.DataFrame = None,
-        data: pd.DataFrame = None,
-        studies: pd.DataFrame = None,
+        timecourses: pd.DataFrame = None,
+        scatters: pd.DataFrame = None,
     ):
         """Creates PKDB data object from given DataFrames.
 
@@ -179,27 +241,71 @@ class PKData(object):
         :param individuals:
         :param groups:
         :param outputs:
-        :param data:
+        :param timecourses:
+        :param scatters:
+
+
         """
+        self.studies = PKDataFrame(studies, pk="sid")
+
         self.groups = PKDataFrame(groups, pk="group_pk")
         self.individuals = PKDataFrame(individuals, pk="individual_pk")
-        self.interventions = PKDataFrame(interventions, pk="intervention_pk")
+        self.interventions = PKDataFrame(interventions, pk="intervention_pk").replace(
+            {np.nan: None}
+        )
         self.outputs = PKDataFrame(outputs, pk="output_pk")
-        self.data = PKDataFrame(data, pk="subset_pk")
-        self.studies = PKDataFrame(studies, pk="sid")
+        self.timecourses = PKDataFrame(timecourses, pk="subset_pk")
+        self.scatters = PKDataFrame(scatters, pk="subset_pk")
 
         if not self.individuals.empty:
             self.individuals.substance = self.individuals.substance.astype(str)
         if not self.groups.empty:
             self.groups.substance = self.groups.substance.astype(str)
 
+        def _transform_strings_tuple(value):
+            if isinstance(value, str):
+                if value.startswith("[") or value.startswith("("):
+                    return tuple(np.fromstring(value[1:-1], sep=", "))
+                    # return tuple([z for z in value[1:-1].split(",")])
+            return value
+
+        if not self.timecourses.empty:
+            self.timecourses[
+                [
+                    "output_pk",
+                    "intervention_pk",
+                    "mean",
+                    "value",
+                    "sd",
+                    "se",
+                    "min",
+                    "max",
+                ]
+            ] = self.timecourses[
+                [
+                    "output_pk",
+                    "intervention_pk",
+                    "mean",
+                    "value",
+                    "sd",
+                    "se",
+                    "min",
+                    "max",
+                ]
+            ].df.applymap(
+                _transform_strings_tuple
+            )
+
     def __dict___(self):
+        """serialises pkdata instance to a dict."""
         return {df_key: getattr(self, df_key).df for df_key in PKData.KEYS}
 
     def as_dict(self):
+        """serialises pkdata instance to a dict."""
         return self.__dict___()
 
     def copy(self):
+        """creates a copy of the pkdata instance."""
         return PKData(**self.as_dict())
 
     def __str__(self):
@@ -221,7 +327,7 @@ class PKData(object):
                 lines.append(f"{key:<15} {count:>5}  ({nrows:>5})")
             else:
                 lines.append(
-                    f"{key:<12} tc:{ self.timecourses_count:>5}  sc:{self.scatter_count:>3}"
+                    f"{key:<12} tc:{self.timecourses_count:>5}  sc:{self.scatter_count:>3}"
                 )
 
         lines.append("-" * 30)
@@ -275,9 +381,40 @@ class PKData(object):
 
         return PKData(**resulting_kwargs)
 
+    @classmethod
+    def from_download(cls, path: Union[BytesIO, os.PathLike]) -> "PKData":
+        """Load data from downloaded zip archive."""
+        pkdata = cls.from_archive(path=path)
+        # fix the intervention keys due to different serialization format
+        pkdata = cls._intervention_pk_update(pkdata)
+        return pkdata
+
+    @classmethod
+    def from_archive(cls, path: Union[BytesIO, os.PathLike]) -> "PKData":
+        """Load data from serialized archive."""
+        data_dict = {}
+        with zipfile.ZipFile(path, "r") as archive:
+            for key in PKData.KEYS:
+                df = pd.read_csv(archive.open(f"{key}.csv", "r"), low_memory=False)
+                data_dict[key] = PKData._clean_types(
+                    df, is_array=key in ["timecourses", "scatters"]
+                )
+        # create data from data frames
+        return PKData(**data_dict)
+
+    def to_archive(self, path: Path) -> None:
+        """Saves data to zip archive"""
+        create_parent(path)
+        with zipfile.ZipFile(path, "w") as archive:
+            for key in PKData.KEYS:
+                df = getattr(self, key)  # type: pd.DataFrame
+                with tempfile.NamedTemporaryFile() as fp:
+                    df.to_csv(fp.name)
+                    archive.write(filename=fp.name, arcname=f"{key}.csv")
+
     @staticmethod
     def from_hdf5(path: Path) -> "PKData":
-        """Load data from HDF5 serialization.
+        """Load data from an archive as returned from the download in pk-db.com.
 
         :param path: path to HDF5.
         :type path: str
@@ -294,11 +431,8 @@ class PKData(object):
         return PKData(**data_dict)
 
     def to_hdf5(self, path: Path) -> None:
-        """Saves data HDF5
-
-        :param path: path to HDF5.
-        :type path: str
-        """
+        """Saves data HDF5."""
+        create_parent(path)
         store = pd.HDFStore(path)
         for key in [
             "studies",
@@ -306,7 +440,7 @@ class PKData(object):
             "individuals",
             "groups",
             "outputs",
-            "data",
+            "timecourses",
         ]:
             df = getattr(self, key).df
             store[key] = df
@@ -326,6 +460,12 @@ class PKData(object):
 
         return study_sids
 
+    def healthy(self):
+        """subset of healthy data."""
+        return self.filter_subject(f_healthy, concise=False).exclude_subject(
+            f_n_healthy
+        )
+
     @property
     def groups_count(self) -> int:
         """Number of groups contained in this PKData instance.
@@ -369,7 +509,7 @@ class PKData(object):
         :return: Number of timecourses contained in this PKData instance.
         :rtype: int
         """
-        return self.data[self.data.data_type == "timecourse"].pk_len
+        return self.timecourses.pk_len
 
     @property
     def scatter_count(self) -> int:
@@ -382,7 +522,7 @@ class PKData(object):
 
     @property
     def timecourses_extended(self) -> pd.DataFrame:
-        """ extends the timecourse df with the core information from interventions, individuals and groups"""
+        """extends the timecourse df with the core information from interventions, individuals and groups"""
 
         timecourses = self.timecourses.df.merge(
             self.interventions_core,
@@ -496,6 +636,17 @@ class PKData(object):
             "outputs", ["output_pk", "intervention_pk", "group_pk", "individual_pk"]
         )
 
+    @property
+    def timecourses_mi(self) -> pd.DataFrame:
+        """Multi-index DataFrame of timecourses contained in this PKData instance.
+
+        :return: Multi-indexed DataFrame of timecourses contained in this PKData instance.
+        :rtype: pd.DataFrame
+        """
+        return self._df_mi(
+            "timecourses", ["subset_pk", "intervention_pk", "group_pk", "individual_pk"]
+        )
+
     # --- filter and exclude ---
     def _pk_filter(
         self, df_key: str, f_idx, concise: bool, *args, **kwargs
@@ -515,16 +666,17 @@ class PKData(object):
         return pkdata
 
     def _pk_exclude(self, df_k, f_idx, concise, **kwargs) -> "PKData":
+        """Generic function to exclude data selected by the table key (df_k) and filtered by f_idx."""
 
         dict_pkdata = self.as_dict()
         dict_pkdata[df_k] = getattr(self, df_k).pk_exclude(f_idx, **kwargs)
-
         pkdata = PKData(**dict_pkdata)
         if concise:
             pkdata._concise()
         return pkdata
 
     def _emptify(self, df_key, concise=True) -> "PKData":
+        """generic function to emptify a table selected by the key."""
         self._validate_df_key(df_key)
 
         dict_pkdata = self.as_dict()
@@ -537,13 +689,14 @@ class PKData(object):
 
     @staticmethod
     def _validate_df_key(df_key):
+        """correct key validations function"""
         if df_key not in PKData.KEYS:
             raise ValueError(
                 f"Unsupported key '{df_key}', key must be in '{PKData.KEYS}'"
             )
 
     def filter_study(self, f_idx, concise=True, **kwargs) -> "PKData":
-        """ Filter studies by filter function. """
+        """Filter studies by filter function."""
         return self._pk_filter("studies", f_idx, concise, **kwargs)
 
     def filter_intervention(self, f_idx, concise=True, *args, **kwargs) -> "PKData":
@@ -569,16 +722,17 @@ class PKData(object):
         return self._pk_filter("interventions", f_idx, concise, **kwargs)
 
     def filter_group(self, f_idx, concise=True, **kwargs) -> "PKData":
-        """ Filter groups. """
+        """Filter groups."""
+
         return self._pk_filter("groups", f_idx, concise, **kwargs)
 
     def filter_individual(self, f_idx, concise=True, **kwargs) -> "PKData":
+        """Filter individuals."""
 
-        """ Filter individuals. """
         return self._pk_filter("individuals", f_idx, concise, **kwargs)
 
     def filter_subject(self, f_idx, concise=True, **kwargs) -> "PKData":
-        """ Filter group or individual. """
+        """Filter group or individual."""
         pkdata = self.filter_group(f_idx, concise=False, **kwargs)
         pkdata = pkdata.filter_individual(f_idx, concise=False, **kwargs)
         if concise:
@@ -586,27 +740,58 @@ class PKData(object):
         return pkdata
 
     def filter_output(self, f_idx, concise=True, **kwargs) -> "PKData":
-        """ Filter outputs. """
+        """Filter outputs."""
         return self._pk_filter("outputs", f_idx, concise, **kwargs)
 
-    def filter_data(self, f_idx, concise=True, **kwargs) -> "PKData":
-        """ Filter data. """
-        return self._pk_filter("data", f_idx, concise, **kwargs)
+    def filter_timecourse(self, f_idx, concise=True, **kwargs) -> "PKData":
+        """Filter timecourses."""
+        return self._pk_filter("timecourses", f_idx, concise, **kwargs)
+
+    def filter(self, filter_dict: Dict) -> "PKData":
+        pkdata = self.copy()
+        filter_functions = [
+            "groups",
+            "individuals",
+            "interventions",
+            "outputs",
+        ]
+        for key in filter_functions:
+            # filter each table separately by dedicated function
+            table_filter_definitions = filter_dict.get(key, None)
+            if table_filter_definitions:
+                table_filters = filter_factory(table_filter_definitions)
+                pkdata = pkdata._pk_filter(key, f_idx=table_filters, concise=False)
+        pkdata._concise()
+        return pkdata
 
     def exclude_study(self, f_idx, concise=True, **kwargs) -> "PKData":
-        """ Filter groups. """
+        """Excludes studies which cann be selected by a filter (idx)."""
+
         return self._pk_exclude("studies", f_idx, concise, **kwargs)
 
     def exclude_intervention(self, f_idx, concise=True, **kwargs) -> "PKData":
+        """Excludes interventions which cann be selected by a filter (idx)."""
+
         return self._pk_exclude("interventions", f_idx, concise, **kwargs)
 
     def exclude_group(self, f_idx, concise=True, **kwargs) -> "PKData":
+        """Excludes groups which cann be selected by a filter (idx)."""
+
         return self._pk_exclude("groups", f_idx, concise, **kwargs)
 
     def exclude_individual(self, f_idx, concise=True, **kwargs):
+        """Excludes individuals which cann be selected by a filter (idx)."""
+
         return self._pk_exclude("individuals", f_idx, concise, **kwargs)
 
     def exclude_subject(self, f_idx, concise=True, **kwargs) -> "PKData":
+        """Excludes groups and individuals which cann be selected by a filter (idx).
+
+        :param f_idx:
+        :param concise:
+        :param kwargs:
+        :return:
+        """
         pkdata = self.exclude_group(f_idx, concise=False, **kwargs)
         pkdata = pkdata.exclude_individual(f_idx, concise=False, **kwargs)
         if concise:
@@ -616,9 +801,9 @@ class PKData(object):
     def exclude_output(self, f_idx, concise=True, **kwargs) -> "PKData":
         return self._pk_exclude("outputs", f_idx, concise, **kwargs)
 
-    def exclude_data(self, f_idx, concise=True, **kwargs):
+    def exclude_timecourse(self, f_idx, concise=True, **kwargs):
 
-        return self._pk_exclude("data", f_idx, concise, **kwargs)
+        return self._pk_exclude("timecourses", f_idx, concise, **kwargs)
 
     def delete_groups(self, concise=True) -> "PKData":
         """
@@ -641,80 +826,79 @@ class PKData(object):
         """
         return self._emptify("outputs", concise=concise)
 
-    def delete_data(self, concise=True) -> "PKData":
-        """Deletes data."""
-        return self._emptify("data", concise=concise)
+    def delete_timecourses(self, concise=True) -> "PKData":
+        """Deletes timecourses."""
+        return self._emptify("timecourses", concise=concise)
+
+    @property
+    def ids(self):
+        """unique ids of all tables within a pkdata instance."""
+        return {
+            "studies": list(self.studies.pks),
+            "groups": list(self.groups.pks),
+            "individuals": list(self.individuals.pks),
+            "interventions": list(self.interventions.pks),
+            "outputs": list(self.outputs.pks),
+            "timecourses": list(self.timecourses.pks),
+            # "scatters": list(self.scatters.pks),
+        }
 
     def _concise(self) -> None:
         """Reduces the current PKData to a consistent subset.
         Modifies the DataFrame in place.
         :return:
         """
-        previous_len = np.inf
-        logger.warning("Concise DataFrames")
-        while previous_len > self._len_total:
-            previous_len = copy(self._len_total)
+        # FIXME: scatters are not concised !!!
 
-            # concise based on studies
-            outputs_study_sids = set(self.outputs.study_sid)
-            data_study_sids = set(self.data.study_sid)
-            interventions_study_sids = set(self.interventions.study_sid)
-            groups_study_sids = set(self.groups.study_sid)
-            individuals_study_sids = set(self.individuals.study_sid)
+        self.outputs = self.outputs[
+            self.outputs["group_pk"].isin(self.ids["groups"])
+            | self.outputs["individual_pk"].isin(self.ids["individuals"])
+        ]
+        self.outputs = self.outputs[
+            self.outputs["intervention_pk"].isin(self.ids["interventions"])
+        ]
 
-            study_sid_sets = [
-                outputs_study_sids,
-                data_study_sids,
-                interventions_study_sids,
-                groups_study_sids,
-                individuals_study_sids,
-            ]
+        concised_ids = {
+            "studies": list(self.outputs.study_sid.unique()),
+            "groups": list(self.outputs.group_pk.unique()),
+            "individuals": list(self.outputs.individual_pk.unique()),
+            "interventions": list(self.outputs.intervention_pk.unique()),
+            "outputs": list(self.outputs.pks),
+            "timecourses": list(self.timecourses.pks),
+            # "scatters": list(self.scatters.pks),
+        }
 
-            current_study_sid_sets = set().union(*study_sid_sets)
-            self.studies = self.studies[self.studies.sid.isin(current_study_sid_sets)]
-            for df_key in ["interventions", "groups", "individuals", "data", "outputs"]:
-                df = getattr(self, df_key)
-                setattr(self, df_key, df[df["study_sid"].isin(self.studies.pks)])
+        self.studies = self.studies[self.studies["sid"].isin(concised_ids["studies"])]
+        self.interventions = self.interventions[
+            self.interventions["intervention_pk"].isin(concised_ids["interventions"])
+        ]
+        self.groups = self.groups[self.groups["group_pk"].isin(concised_ids["groups"])]
+        self.individuals = self.individuals[
+            self.individuals["individual_pk"].isin(concised_ids["individuals"])
+        ]
 
-            # concise based on interventions
-            outputs_intervention_pks = set(self.outputs.intervention_pk)
-            current_intervention_pks = self.interventions.pks.intersection(
-                outputs_intervention_pks
+        if not self.timecourses.empty:
+            _timecourses = pd.DataFrame(
+                {
+                    "subset_pk": np.repeat(
+                        self.timecourses.subset_pk.values,
+                        self.timecourses.output_pk.str.len(),
+                    ),
+                    "output_pk": np.concatenate(self.timecourses.output_pk.values),
+                }
             )
-            current_intervention_pks.add(-1)
+            _timecourses["output_pk"] = _timecourses["output_pk"].astype(int)
 
-            self.interventions = self.interventions[
-                self.interventions.intervention_pk.isin(current_intervention_pks)
+            _timecourses = _timecourses[
+                _timecourses["output_pk"].isin(concised_ids["outputs"])
             ]
-            self.outputs = self.outputs[
-                self.outputs.intervention_pk.isin(current_intervention_pks)
+            self.timecourses = self.timecourses[
+                self.timecourses["subset_pk"].isin(_timecourses.subset_pk.unique())
             ]
 
-            # concise based on individuals
-            outputs_individual_pks = set(self.outputs.individual_pk)
-            current_individual_pks = self.individuals.pks.intersection(
-                outputs_individual_pks
-            )
-            current_individual_pks.add(-1)
-
-            for df_key in ["individuals", "outputs"]:
-                df = getattr(self, df_key)
-                setattr(
-                    self, df_key, df[df["individual_pk"].isin(current_individual_pks)]
-                )
-
-            # concise based on groups
-            outputs_group_pks = set(self.outputs.group_pk)
-
-            current_group_pks = self.groups.pks.intersection(outputs_group_pks)
-            current_group_pks.add(-1)
-
-            for df_key in ["groups", "outputs"]:
-                df = getattr(self, df_key)
-                setattr(self, df_key, df[df.group_pk.isin(current_group_pks)])
-
     @property
     def _len_total(self):
+        """The sum of all entries in all tables."""
         return sum([len(getattr(self, df_key)) for df_key in PKData.KEYS])
 
     def get_choices(self):
@@ -764,3 +948,178 @@ class PKData(object):
             for field in fields:
                 print(f"*** {field} ***")
                 print(choices[field])
+
+    def _map_intervention_pks(self):
+        """Helper Function for the transformation of intervention_pk in outputs and timecourses.
+        returns a mapping of old intervetions_pks with new intervetion_pks
+        """
+        interventions_output = pd.DataFrame()
+
+        if not self.outputs.empty:
+            interventions_output = self.outputs.df.pivot_table(
+                values="intervention_pk",
+                index="output_pk",
+                aggfunc=lambda x: frozenset(x),
+            )
+
+        interventions = interventions_output.drop_duplicates(
+            "intervention_pk"
+        ).reset_index()
+        interventions.index = interventions.index.set_names(["intervention_pk_updated"])
+        return interventions["intervention_pk"].reset_index()
+
+    def _update_interventions(self, mapping_int_pks):
+        """Updates intervention_pk based on if they are beeing used in the outputs. Multiple interventions can have the same
+        intervention_pk. After the transformation each (output) row in the outputs links only to one intervention_pk."""
+        """FIXME: document me"""
+        mapping_int_pks = mapping_int_pks.copy()
+        mapping_int_pks["intervention_pk"] = mapping_int_pks.intervention_pk.apply(
+            lambda x: list(x)
+        )
+        mapping_int_pks = (
+            mapping_int_pks.intervention_pk.apply(pd.Series)
+            .stack()
+            .reset_index(level=-1, drop=True)
+            .astype(int)
+            .reset_index()
+        )
+        mapping_int_pks = mapping_int_pks.rename(
+            columns={"index": "intervention_pk_updated", 0: "intervention_pk"}
+        )
+        return (
+            pd.merge(mapping_int_pks, self.interventions, on="intervention_pk")
+            .drop(columns=["intervention_pk"])
+            .rename(columns={"intervention_pk_updated": "intervention_pk"})
+        )
+
+    def _update_outputs(self, mapping_int_pks):
+        """Dates up all intervention_pk in outputs table. Thereby each row becomes a unique output."""
+        mapping_int_pks = mapping_int_pks.copy()
+
+        interventions_output = self.outputs.df.pivot_table(
+            values="intervention_pk", index="output_pk", aggfunc=lambda x: frozenset(x)
+        )
+        mapping_int_pks = pd.merge(
+            interventions_output.reset_index(),
+            mapping_int_pks,
+            on="intervention_pk",
+            how="left",
+        )[["output_pk", "intervention_pk_updated"]]
+
+        return (
+            pd.merge(
+                mapping_int_pks,
+                self.outputs.df.drop_duplicates(subset="output_pk"),
+                how="left",
+            )
+            .drop(columns=["intervention_pk"])
+            .rename(columns={"intervention_pk_updated": "intervention_pk"})
+        )
+
+    def get_updated_intervention_pk(self, frozenset_intervention_pks):
+        """return new set"""
+
+    def _update_timecourses(self, mapping_int_pks):
+        """Dates up all intervention_pk in timecourses table."""
+        mapping_dict = (
+            mapping_int_pks.copy()
+            .set_index("intervention_pk")["intervention_pk_updated"]
+            .to_dict()
+        )
+        self.timecourses["intervention_pk"] = self.timecourses["intervention_pk"].apply(
+            lambda x: mapping_dict.get(frozenset(x))
+        )
+
+        return self.timecourses
+
+    def _update_scatters(self, mapping_int_pks):
+        """Dates up all intervention_pk in scatters table."""
+        mapping_dict = (
+            mapping_int_pks.copy()
+            .set_index("intervention_pk")["intervention_pk_updated"]
+            .to_dict()
+        )
+
+        self.scatters["x_intervention_pk"] = self.scatters["x_intervention_pk"].apply(
+            frozenset
+        )
+        self.scatters["y_intervention_pk"] = self.scatters["y_intervention_pk"].apply(
+            frozenset
+        )
+
+        self.scatters["x_intervention_pk"] = self.scatters["x_intervention_pk"].apply(
+            lambda x: mapping_dict.get(frozenset(x))
+        )
+
+        self.scatters["y_intervention_pk"] = self.scatters["y_intervention_pk"].apply(
+            lambda x: mapping_dict.get(frozenset(x))
+        )
+
+    def _intervention_pk_update(self):
+        """Performs all three function necessary to update the
+        intervention pks in all three tables where they are contained (interventions, output, timecourses)."""
+        if self.outputs.empty:
+            return self
+        else:
+            mapping_int_pks = self._map_intervention_pks()
+            data_dict = self.as_dict()
+            data_dict["interventions"] = self._update_interventions(mapping_int_pks)
+            if not self.outputs.empty:
+                data_dict["outputs"] = self._update_outputs(mapping_int_pks)
+            if not self.timecourses.empty:
+                data_dict["timecourses"] = self._update_timecourses(mapping_int_pks)
+            # if not self.scatters.empty:
+            #    data_dict["scatters"] = self._update_scatters(mapping_int_pks)
+            return PKData(**data_dict)
+
+    @staticmethod
+    def _clean_types(df: pd.DataFrame, is_array):
+        """Sets the correct datatypes for each column in the table (df)."""
+        # convert columns to float columns
+        float_columns = [
+            "mean",
+            "median",
+            "value",
+            "sd",
+            "se",
+            "cv",
+            "min",
+            "max",
+            # "time",
+        ]
+        # convert columns to int columns
+        int_columns = [
+            "subset_pk",
+            "group_pk",
+            "individual_pk",
+            "group_parent_pk",
+            "raw_pk",
+        ]
+
+        if not is_array:
+            int_columns.append("intervention_pk")
+            for column in float_columns:
+                if column in df.columns:
+                    df[column] = df[column].astype(float)
+
+        for column in int_columns:
+            if column in df.columns:
+                df[column] = df[column].replace({np.nan: -1}).astype(int)
+
+        return df
+
+    def to_medline(self, path: Path):
+        """create a bibtex file."""
+
+        create_parent(path)
+        reference_pmids = [str(int(s)) for s in self.studies.reference_pmid if s]
+        reference_pmids_str = "%2C".join(reference_pmids)
+        url = (
+            "https://api.ncbi.nlm.nih.gov/lit/ctxp/v1/pubmed/?format=medline&id="
+            + reference_pmids_str
+            + "&download=y"
+        )
+        with requests.get(url) as r:
+            r.raise_for_status()
+            with open(path, "wb") as f:
+                f.write(r.content)
