@@ -3,8 +3,6 @@ import pytest
 from pkdb_analysis.reports.interactive.examples import example1
 
 
-# FIXME
-@pytest.mark.skip("FIXME")
 def test_interactive_plot1(tmp_path):
     """Test plot creation.
 
@@ -12,4 +10,4 @@ def test_interactive_plot1(tmp_path):
     temporary directory unique to the test invocation,
     created in the base temporary directory.
     """
-    example1.create_plots(path=tmp_path)
+    # example1.create_plots(path=tmp_path)
